        .arch at90s1200
;#define at90s1200
;
; Hardware
;
; Define the hardware interface and bits
;
; Revision History
;       3/2/2001 - Added bit to enable pull-up on PD6, Brake enable.
;
#include "at90s1200.inc"

#define BV(A) (1<<(A))

#define CPUCLK 16000000L
;
; Tickrate is to be 250k for a 4us resolution on
; the R/C Pulse input.
;
#define TCCR0_INIT CLK64	; ck/64
#define TCNT0_INIT 0xFF         ; Single tick/interrupt.
#define	TIMSK_INIT (1<<TOIE0)
#define MCUCR_INIT (1<<SE)      // Enable sleep mode
;
; PWM output is +/- 75 for a PWM rate of ~1.7khz and R/C pulse range of 600us
; PWM output is +/- 125 for a PWM rate of 1.0khz and R/C pulse range of 1000us
;
#define PWM_600US       75      // Sets range of PWM input (+/- 75 * 4us or .6ms)
#define PWM_1000US      125     // Sets range of PWM input (+/- 125 * 4us or 1.0ms)
#define RC_CENTER       (-375)  // 1.5ms/4us
#define RC_PULSEFAIL    100     // ~60 ms for pulse failure detact
;
; Port pin definitions
;
#define BRAKE    PIND, PD6      // 1 = coast, 0 = Brake
#define FWD_A   PORTD, PD5      // Channel A motor driver signals
#define REV_A   PORTD, PD4
#define ENA_A   PORTD, PD3
#define FWD_B   PORTD, PD2      // Channel B Motor Driver Signals
#define REV_B   PORTD, PD1
#define ENA_B   PORTD, PD0

#define DDRD_INIT 0x3F          // PD6 in, PD5-0 out
#define PORTD_INIT 0x40         // Pullup on PD6

#define DB_1     PINB, PB7      // Deadband select: +/- 0, 4, 8 or 12us
#define MIXED    PINB, PB6      // Mixed vs separate
#define NotUsed  PINB, PB5
#define RANGE    PINB, PB4      // 1 = .6 ms, 0 = 1.0 ms R/C pulse range
#define RC_A_INP PINB, PB3      // A channel R/C pulse input
#define SHDWN_A  PINB, PB2      // 0 = Kill output channel A
#define RC_B_INP PINB, PB1      // B channel R/C pulse input
#define SHDWN_B  PINB, PB0      // 0 = Kill output channel B

#define DDRB_INIT 0x00          // All Inputs with pullups
#define PORTB_INIT 0xFF
;
; Register Definitions
;
#define zero            R1

#define a_pwm           R3      // Working PWM values
#define b_pwm           R4
#define centerl         R5      // Center value for 1.5ms pulse
#define centerh         R6
#define p1              R7      // Passed Parameter 1
#define rp1             R8      // Returned Parameter 1
#define p2              R8
#define b_hold          R9      // Holding registers for last computed A & B Pulse width
#define a_hold          R10
#define pwm_dir         R11

#define sreg_save       R16
#define range           R17
#define deadband        R18     // Deadband in 4us increments
#define count           R19     // PWM Counter
#define flags           R20
#define     RC_A_ST         0   // Bit Values
#define     RC_A_END        1
#define     RC_B_ST         2
#define     RC_B_END        3
#define     PWM_DIR         4
#define swtmp           R21     // High register for swread use...

#define rc_a_cntl       R22     // R/C Pulse width counters
#define rc_a_cnth       R23
#define rc_b_cntl       R24
#define rc_b_cnth       R25

#define a_pulsefail     R28     // Pulse fail detect counters
#define b_pulsefail     R29
;
; Init section of code with reset and interrupt vectors.
;
; Note: Compiling for the 2313 will work for the 1200 because unprogrammed
; locations are all NOPs. So, regardless which vector is jumped to, the processor
; will eventually find the overfloww vector.
;
        .section        .init
        .func   _init
_init:
        rjmp    reset
        rjmp    halt
#ifdef at90s1200
        .org    0x4
#else
        .org    0xC
#endif
        rjmp    _overflow0_
halt:
        cli
        rjmp    .
        
        .endfunc
        
	.section        .text
;
; main routine
;
        .func   reset
        .global reset
reset:
        ; Initialize all CPU hardware peripherals
        ;
	ldi	R16, PORTD_INIT	; set output state before direction.
	out	PORTD, R16		; at reset output is low, direction is in.
	ldi	R16, DDRD_INIT
	out	DDRD, R16
	ldi	R16, PORTB_INIT	; ditto
	out	PORTB, R16
	ldi	R16, DDRB_INIT
	out	DDRB, R16

	ldi	R16, TCNT0_INIT
	out	TCNT0, R16
	ldi	R16, TCCR0_INIT
	out	TCCR0, R16
	ldi	R16, TIMSK_INIT	; yeah, this is just a bit set...
	out	TIMSK, R16

	ldi	R16, MCUCR_INIT	; set up sleep mode & Extn interrupt
	out	MCUCR, R16
	;
	; Initialize all working registers
        clr     zero
        clr     flags
        clr     count
        clr     a_pwm
        clr     b_pwm
        clr     a_pulsefail
        clr     b_pulsefail
        ldi     R16, 1
        mov     pwm_dir, r16    // Initialize counting up
        ;
        ; Initalize counters so first pass doesnt spike the motors
        ;
        ldi     R16, lo8(RC_CENTER)
        mov     centerl, R16
        mov     rc_a_cntl, R16
        mov     rc_b_cntl, R16
        ldi     R16, hi8(RC_CENTER)
        mov     centerh, R16
        mov     rc_a_cnth, R16
        mov     rc_b_cnth, R16

#ifndef at90s1200
        out     0x3C, zero              ; Init stack for non-1200 chips
        ldi     R16, 0xDF
        out     0x3D, R16
#endif
        sei
        
forever:
	sleep				; Wait for an interrupt
	tst     count
	brne    forever
	rcall   sw_read                 ; Analyze data every PWM rollover cycle
	rjmp	forever
       
        .endfunc
;
; Timer Interrupt handler
;
; Reload timer, examine inputs, drive outputs, etc.
; 20-38 cycles/64 (less than 50% load)
;
	.func   _overflow0_
_overflow0_:
	ldi	sreg_save, TCNT0_INIT
	out	TCNT0, sreg_save
	in      sreg_save, SREG
;
; Logic to measure the pulse width of the RC input
; called every 4us.
;
; 10-22 cycles (mostly 10)
;
rc_read:
        sbis    RC_A_INP
        rjmp    1f
        subi    rc_a_cntl, lo8(-1)
        sbci    rc_a_cnth, hi8(-1)      ; add 1 
        sbr     flags, BV(RC_A_ST)
        rjmp    2f
1:
        sbrc    flags, RC_A_ST
        sbr     flags, BV(RC_A_END)
2:
        sbis    RC_B_INP
        rjmp    1f
        subi    rc_b_cntl, lo8(-1)
        sbci    rc_b_cnth, hi8(-1)
        sbr     flags, BV(RC_B_ST)
        rjmp    2f
1:
        sbrc    flags, RC_B_ST
        sbr     flags, BV(RC_B_END)
2:
;
; pwm_control
;
; Counts up/down between 0 and MAXCOUNT
; compares pwm values with count, setting/clearing
; the ENA bit as needed.
; 4/5 cycles
;
pw_control:
        add     count, pwm_dir
        breq    pw_flip
        cp      count, range            ; 75 or 125 (so it *can* be signed)
        brlt    pw_done
pw_flip:
        neg     pwm_dir                 ; flip sense of count
pw_done:
        ;
        ; Set/clear output based upon PWM value and state of Shutdown pin
        ; 11 cycles
        cp      count, a_pwm
        sbic    SHDWN_A                 ; Skip to OFF if shutdown low
        brlo    pw_A_on
pw_A_off:
        cbi     ENA_A
        rjmp    1f
pw_A_on:
        sbi     ENA_A
1:
        cp      count, b_pwm
        sbic    SHDWN_B
        brlo    pw_B_on
pw_B_off:
        cbi     ENA_B
        rjmp    1f
pw_B_on:
        sbi     ENA_B
1:
        out     SREG, sreg_save
        reti
        .endfunc

        .func   sw_read
;
; sw_read
;
; Analyse the RC pulse input and set PWM output.
; Analyse the mode switch input and do calibration modes.
; do pulse failure detect, set output to zero if pulse not present.
;
sw_read:
        ;
        ; Deadband is set by an input pin
        ;
        clr     deadband
        sbic    DB_1
        ldi     deadband, 0x03          ; 12 uS (three ticks)
        ;
        ; Range is set by an input pin
        ;
        sbic    RANGE                   ; Test against range...
        ldi     range, lo8(PWM_600US)
        sbis    RANGE
        ldi     range, lo8(PWM_1000US)
        ;
        ; Analyze the A Channel
        ;
        sbrc    flags, RC_A_END
        rjmp    sw_A_pulse
        ;
        ; Check for missing pulse
        ;
        inc     a_pulsefail
        cpi     a_pulsefail, 100        ; 100 x .6 or 1.0 ms till failure
        brlt    sw_check_B
        dec     a_pulsefail
        mov     a_hold, zero
        rjmp    sw_check_B
        ;
        ; Gotta pulse!
        ;
sw_A_pulse:
        clr     a_pulsefail
        cbr     flags, BV(RC_A_ST) | BV(RC_A_END)
        ;
        mov     p1, rc_a_cntl
        mov     p2, rc_a_cnth
        rcall   CheckCount
        mov     a_hold, p1
        mov     rc_a_cntl, centerl
        mov     rc_a_cnth, centerh

sw_check_B:
        sbrc    flags, RC_B_END
        rjmp    sw_B_pulse

        inc     b_pulsefail
        cpi     b_pulsefail, 100
        brge    1f
        rjmp    sw_done
1:
        dec     b_pulsefail
        mov     b_hold, zero
        rjmp    SetPwmValues
sw_B_pulse:
        clr     b_pulsefail
        cbr     flags, BV(RC_B_ST) | BV(RC_B_END)
        ;
        mov     p1, rc_b_cntl
        mov     p2, rc_b_cnth
        rcall   CheckCount
        mov     b_hold, p1
        mov     rc_b_cntl, centerl
        mov     rc_b_cnth, centerh

SetPwmValues:
        sbic    MIXED                   ; If jumpered, then mix
        rjmp    straight
        
        mov     R30, b_hold             ; Sign extend b_hold
        clr     R31
        sbrc    R30, 7
        com     R31

        mov     p1, a_hold           ; Compute mixed signal
        clr     p2
        sbrc    p1, 7
        com     p2
        add     p1, R30
        adc     p2, R31
        rcall   CheckCount           ; Limit to +/- 1/2 the range.
        mov     a_pwm, p1

        mov     p1, a_hold           ; Compute mixed signal
        clr     p2
        sbrc    p1, 7
        com     p2
        sub     p1, R30
        sbc     p2, R31
        rcall   CheckCount
        mov     b_pwm, p1
        rjmp    endmix
straight:
        mov     a_pwm, a_hold           ; Straight across
        mov     b_pwm, b_hold
endmix:
;
; Now, set drive bits, etc and wrap up
;
        tst     a_pwm
        brlt    sw_a_reverse
        breq    sw_a_halt
sw_a_forward:
        sbi     FWD_A
        cbi     REV_A
        rjmp    sw_a_reload
sw_a_reverse:
        cbi     FWD_A
        sbi     REV_A
        neg     a_pwm
        rjmp    sw_a_reload
sw_a_halt:
        cbi     FWD_A
        cbi     REV_A
        sbic    BRAKE
        rjmp    sw_a_reload
        mov     a_pwm, range
        inc     a_pwm
sw_a_reload:
;
; channel B bits
;
        tst     b_pwm
        brlt    sw_b_reverse
        breq    sw_b_halt
sw_b_forward:
        sbi     FWD_B
        cbi     REV_B
        rjmp    sw_b_reload
sw_b_reverse:
        cbi     FWD_B
        sbi     REV_B
        neg     b_pwm
        rjmp    sw_b_reload
sw_b_halt:
        cbi     FWD_B
        cbi     REV_B
        sbic    BRAKE
        rjmp    sw_b_reload
        mov     b_pwm, range    ; force enable high always
        inc     b_pwm           ; when braking.
sw_b_reload:

sw_done:
        ret
        .endfunc
        
        .func   CheckCount
;
; CheckCount
;
; Passed:
;       p1, p2 = low & high byte of count
; Return
;       p1 = signed value of power
; 
CheckCount:
        cp      p1, zero
        cpc     p2, zero
        brlt    reverse
forward:
        cp      deadband, p1
        cpc     zero, p2
        brge    1f
        cp      p1, range
        cpc     p2, zero
        brlt    2f
        mov     p1, range
2:
        tst     p1
        ret
1:
        clr     p1
        ret
reverse:
        ldi     swtmp, 1
        com     p1               ; negate count
        com     p2
        add     p1, swtmp
        adc     p2, zero
        
        cp      deadband, p1
        cpc     zero, p2
        brge    1b
        cp      p1, range
        cpc     p2, zero
        brlt    2f
        mov     p1, range
2:
        neg     p1              ; return neg flags
        ret
	.endfunc

